@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: ogr_core.h 28900 2015-04-14 09:40:34Z rouault $
+ * $Id: ogr_core.h 971ad299681ca1ea2e1b800e88209f426b77e9aa 2018-04-17 12:14:43 +0200 Even Rouault $
  *
  * Project:  OpenGIS Simple Features Reference Implementation
  * Purpose:  Define some core portability services for cross-platform OGR code.
@@ -32,6 +32,9 @@
 #define OGR_CORE_H_INCLUDED
 
 #include "cpl_port.h"
+#if defined(GDAL_COMPILATION)
+#define DO_NOT_DEFINE_GDAL_RELEASE_DATE_AND_GDAL_RELEASE_NAME
+#endif
 #include "gdal_version.h"
 
 /**
@@ -44,11 +47,21 @@
  * Simple container for a bounding region.
  */
 
-#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)
+/*! @cond Doxygen_Suppress */
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS) && !defined(DOXYGEN_SKIP)
+
+extern "C++"
+{
+#include <limits>
+
+// cppcheck-suppress copyCtorAndEqOperator
 class CPL_DLL OGREnvelope
 {
   public:
-        OGREnvelope() : MinX(0.0), MaxX(0.0), MinY(0.0), MaxY(0.0)
+        OGREnvelope() : MinX(std::numeric_limits<double>::infinity()),
+                        MaxX(-std::numeric_limits<double>::infinity()),
+                        MinY(std::numeric_limits<double>::infinity()),
+                        MaxY(-std::numeric_limits<double>::infinity())
         {
         }
 
@@ -62,48 +75,30 @@ class CPL_DLL OGREnvelope
     double      MinY;
     double      MaxY;
 
-/* See http://trac.osgeo.org/gdal/ticket/5299 for details on this pragma */
-#if ((__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)) && !defined(_MSC_VER)) 
+#ifdef HAVE_GCC_DIAGNOSTIC_PUSH
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wfloat-equal"
 #endif
-    int  IsInit() const { return MinX != 0 || MinY != 0 || MaxX != 0 || MaxY != 0; }
+    int  IsInit() const { return MinX != std::numeric_limits<double>::infinity(); }
 
-#if ((__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)) && !defined(_MSC_VER))
+#ifdef HAVE_GCC_DIAGNOSTIC_PUSH
 #pragma GCC diagnostic pop
 #endif
 
     void Merge( OGREnvelope const& sOther ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,sOther.MinX);
-            MaxX = MAX(MaxX,sOther.MaxX);
-            MinY = MIN(MinY,sOther.MinY);
-            MaxY = MAX(MaxY,sOther.MaxY);
-        }
-        else
-        {
-            MinX = sOther.MinX;
-            MaxX = sOther.MaxX;
-            MinY = sOther.MinY;
-            MaxY = sOther.MaxY;
-        }
+        MinX = MIN(MinX,sOther.MinX);
+        MaxX = MAX(MaxX,sOther.MaxX);
+        MinY = MIN(MinY,sOther.MinY);
+        MaxY = MAX(MaxY,sOther.MaxY);
     }
+
     void Merge( double dfX, double dfY ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,dfX);
-            MaxX = MAX(MaxX,dfX);
-            MinY = MIN(MinY,dfY);
-            MaxY = MAX(MaxY,dfY);
-        }
-        else
-        {
-            MinX = MaxX = dfX;
-            MinY = MaxY = dfY;
-        }
+        MinX = MIN(MinX,dfX);
+        MaxX = MAX(MaxX,dfX);
+        MinY = MIN(MinY,dfY);
+        MaxY = MAX(MaxY,dfY);
     }
-    
+
     void Intersect( OGREnvelope const& sOther ) {
         if(Intersects(sOther))
         {
@@ -124,16 +119,13 @@ class CPL_DLL OGREnvelope
         }
         else
         {
-            MinX = 0;
-            MaxX = 0;
-            MinY = 0;
-            MaxY = 0;
+            *this = OGREnvelope();
         }
     }
- 
+
     int Intersects(OGREnvelope const& other) const
     {
-        return MinX <= other.MaxX && MaxX >= other.MinX && 
+        return MinX <= other.MaxX && MaxX >= other.MinX &&
                MinY <= other.MaxY && MaxY >= other.MinY;
     }
 
@@ -143,6 +135,9 @@ class CPL_DLL OGREnvelope
                MaxX >= other.MaxX && MaxY >= other.MaxY;
     }
 };
+
+} // extern "C++"
+
 #else
 typedef struct
 {
@@ -153,16 +148,21 @@ typedef struct
 } OGREnvelope;
 #endif
 
-
 /**
  * Simple container for a bounding region in 3D.
  */
 
-#if defined(__cplusplus) && !defined(CPL_SURESS_CPLUSPLUS)
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS) && !defined(DOXYGEN_SKIP)
+
+extern "C++" {
+
+// cppcheck-suppress copyCtorAndEqOperator
 class CPL_DLL OGREnvelope3D : public OGREnvelope
 {
   public:
-        OGREnvelope3D() : OGREnvelope(), MinZ(0.0), MaxZ(0.0)
+        OGREnvelope3D() : OGREnvelope(),
+                          MinZ(std::numeric_limits<double>::infinity()),
+                          MaxZ(-std::numeric_limits<double>::infinity())
         {
         }
 
@@ -175,43 +175,31 @@ class CPL_DLL OGREnvelope3D : public OGREnvelope
     double      MinZ;
     double      MaxZ;
 
-    int  IsInit() const { return MinX != 0 || MinY != 0 || MaxX != 0 || MaxY != 0 || MinZ != 0 || MaxZ != 0; }
+#ifdef HAVE_GCC_DIAGNOSTIC_PUSH
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wfloat-equal"
+#endif
+    int  IsInit() const { return MinX != std::numeric_limits<double>::infinity(); }
+#ifdef HAVE_GCC_DIAGNOSTIC_PUSH
+#pragma GCC diagnostic pop
+#endif
+
     void Merge( OGREnvelope3D const& sOther ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,sOther.MinX);
-            MaxX = MAX(MaxX,sOther.MaxX);
-            MinY = MIN(MinY,sOther.MinY);
-            MaxY = MAX(MaxY,sOther.MaxY);
-            MinZ = MIN(MinZ,sOther.MinZ);
-            MaxZ = MAX(MaxZ,sOther.MaxZ);
-        }
-        else
-        {
-            MinX = sOther.MinX;
-            MaxX = sOther.MaxX;
-            MinY = sOther.MinY;
-            MaxY = sOther.MaxY;
-            MinZ = sOther.MinZ;
-            MaxZ = sOther.MaxZ;
-        }
+        MinX = MIN(MinX,sOther.MinX);
+        MaxX = MAX(MaxX,sOther.MaxX);
+        MinY = MIN(MinY,sOther.MinY);
+        MaxY = MAX(MaxY,sOther.MaxY);
+        MinZ = MIN(MinZ,sOther.MinZ);
+        MaxZ = MAX(MaxZ,sOther.MaxZ);
     }
+
     void Merge( double dfX, double dfY, double dfZ ) {
-        if( IsInit() )
-        {
-            MinX = MIN(MinX,dfX);
-            MaxX = MAX(MaxX,dfX);
-            MinY = MIN(MinY,dfY);
-            MaxY = MAX(MaxY,dfY);
-            MinZ = MIN(MinZ,dfZ);
-            MaxZ = MAX(MaxZ,dfZ);
-        }
-        else
-        {
-            MinX = MaxX = dfX;
-            MinY = MaxY = dfY;
-            MinZ = MaxZ = dfZ;
-        }
+        MinX = MIN(MinX,dfX);
+        MaxX = MAX(MaxX,dfX);
+        MinY = MIN(MinY,dfY);
+        MaxY = MAX(MaxY,dfY);
+        MinZ = MIN(MinZ,dfZ);
+        MaxZ = MAX(MaxZ,dfZ);
     }
 
     void Intersect( OGREnvelope3D const& sOther ) {
@@ -238,12 +226,7 @@ class CPL_DLL OGREnvelope3D : public OGREnvelope
         }
         else
         {
-            MinX = 0;
-            MaxX = 0;
-            MinY = 0;
-            MaxY = 0;
-            MinZ = 0;
-            MaxZ = 0;
+            *this = OGREnvelope3D();
         }
     }
 
@@ -261,6 +244,9 @@ class CPL_DLL OGREnvelope3D : public OGREnvelope
                MinZ <= other.MinZ && MaxZ >= other.MaxZ;
     }
 };
+
+} // extern "C++"
+
 #else
 typedef struct
 {
@@ -272,29 +258,51 @@ typedef struct
     double      MaxZ;
 } OGREnvelope3D;
 #endif
-
+/*! @endcond */
 
 CPL_C_START
 
-void CPL_DLL *OGRMalloc( size_t );
-void CPL_DLL *OGRCalloc( size_t, size_t );
-void CPL_DLL *OGRRealloc( void *, size_t );
-char CPL_DLL *OGRStrdup( const char * );
-void CPL_DLL OGRFree( void * );
+/*! @cond Doxygen_Suppress */
+void CPL_DLL *OGRMalloc( size_t ) CPL_WARN_DEPRECATED("Use CPLMalloc instead.");
+void CPL_DLL *OGRCalloc( size_t, size_t ) CPL_WARN_DEPRECATED("Use CPLCalloc instead.");
+void CPL_DLL *OGRRealloc( void *, size_t ) CPL_WARN_DEPRECATED("Use CPLRealloc instead.");
+char CPL_DLL *OGRStrdup( const char * ) CPL_WARN_DEPRECATED("Use CPLStrdup instead.");
+void CPL_DLL OGRFree( void * ) CPL_WARN_DEPRECATED("Use CPLFree instead.");
+/*! @endcond */
 
+#ifdef STRICT_OGRERR_TYPE
+/** Type for a OGR error */
+typedef enum
+{
+    OGRERR_NONE,                       /**< Success */
+    OGRERR_NOT_ENOUGH_DATA,            /**< Not enough data to deserialize */
+    OGRERR_NOT_ENOUGH_MEMORY,          /**< Not enough memory */
+    OGRERR_UNSUPPORTED_GEOMETRY_TYPE,  /**< Unsupported geometry type */
+    OGRERR_UNSUPPORTED_OPERATION,      /**< Unsupported operation */
+    OGRERR_CORRUPT_DATA,               /**< Corrupt data */
+    OGRERR_FAILURE,                    /**< Failure */
+    OGRERR_UNSUPPORTED_SRS,            /**< Unsupported SRS */
+    OGRERR_INVALID_HANDLE,             /**< Invalid handle */
+    OGRERR_NON_EXISTING_FEATURE        /**< Non existing feature. Added in GDAL 2.0 */
+} OGRErr;
+#else
+/** Type for a OGR error */
 typedef int OGRErr;
 
-#define OGRERR_NONE                0
-#define OGRERR_NOT_ENOUGH_DATA     1    /* not enough data to deserialize */
-#define OGRERR_NOT_ENOUGH_MEMORY   2
-#define OGRERR_UNSUPPORTED_GEOMETRY_TYPE 3
-#define OGRERR_UNSUPPORTED_OPERATION 4
-#define OGRERR_CORRUPT_DATA        5
-#define OGRERR_FAILURE             6
-#define OGRERR_UNSUPPORTED_SRS     7
-#define OGRERR_INVALID_HANDLE      8
-#define OGRERR_NON_EXISTING_FEATURE 9   /* added in GDAL 2.0 */
+#define OGRERR_NONE                0       /**< Success */
+#define OGRERR_NOT_ENOUGH_DATA     1       /**< Not enough data to deserialize */
+#define OGRERR_NOT_ENOUGH_MEMORY   2       /**< Not enough memory */
+#define OGRERR_UNSUPPORTED_GEOMETRY_TYPE 3 /**< Unsupported geometry type */
+#define OGRERR_UNSUPPORTED_OPERATION 4     /**< Unsupported operation */
+#define OGRERR_CORRUPT_DATA        5       /**< Corrupt data */
+#define OGRERR_FAILURE             6       /**< Failure */
+#define OGRERR_UNSUPPORTED_SRS     7       /**< Unsupported SRS */
+#define OGRERR_INVALID_HANDLE      8       /**< Invalid handle */
+#define OGRERR_NON_EXISTING_FEATURE 9      /**< Non existing feature. Added in GDAL 2.0 */
 
+#endif
+
+/** Type for a OGR boolean */
 typedef int     OGRBoolean;
 
 /* -------------------------------------------------------------------- */
@@ -306,7 +314,7 @@ typedef int     OGRBoolean;
  * but are also returned from OGRGeometry::getGeometryType() to identify the
  * type of a geometry object.
  */
-typedef enum 
+typedef enum
 {
     wkbUnknown = 0,         /**< unknown type, non-standard */
 
@@ -330,6 +338,13 @@ typedef enum
                              *    ISO SQL/MM Part 3. GDAL &gt;= 2.0 */
     wkbMultiCurve = 11,     /**< GeometryCollection of Curves, ISO SQL/MM Part 3. GDAL &gt;= 2.0 */
     wkbMultiSurface = 12,   /**< GeometryCollection of Surfaces, ISO SQL/MM Part 3. GDAL &gt;= 2.0 */
+    wkbCurve = 13,          /**< Curve (abstract type). ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbSurface = 14,        /**< Surface (abstract type). ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbPolyhedralSurface = 15,/**< a contiguous collection of polygons, which share common boundary segments,
+                               *   ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTIN = 16,              /**< a PolyhedralSurface consisting only of Triangle patches
+                               *    ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTriangle = 17,         /**< a Triangle. ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
 
     wkbNone = 100,          /**< non-standard, for pure attribute records */
     wkbLinearRing = 101,    /**< non-standard, just for createGeometry() */
@@ -339,6 +354,47 @@ typedef enum
     wkbCurvePolygonZ = 1010,    /**< wkbCurvePolygon with Z component. ISO SQL/MM Part 3. GDAL &gt;= 2.0 */
     wkbMultiCurveZ = 1011,      /**< wkbMultiCurve with Z component. ISO SQL/MM Part 3. GDAL &gt;= 2.0 */
     wkbMultiSurfaceZ = 1012,    /**< wkbMultiSurface with Z component. ISO SQL/MM Part 3. GDAL &gt;= 2.0 */
+    wkbCurveZ = 1013,           /**< wkbCurve with Z component. ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbSurfaceZ = 1014,         /**< wkbSurface with Z component. ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbPolyhedralSurfaceZ = 1015,  /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTINZ = 1016,                /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTriangleZ = 1017,           /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+
+    wkbPointM = 2001,              /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbLineStringM = 2002,         /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbPolygonM = 2003,            /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiPointM = 2004,         /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiLineStringM = 2005,    /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiPolygonM = 2006,       /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbGeometryCollectionM = 2007, /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCircularStringM = 2008,     /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCompoundCurveM = 2009,      /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCurvePolygonM = 2010,       /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiCurveM = 2011,         /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiSurfaceM = 2012,       /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCurveM = 2013,              /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbSurfaceM = 2014,            /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbPolyhedralSurfaceM = 2015,  /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTINM = 2016,                /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTriangleM = 2017,           /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+
+    wkbPointZM = 3001,              /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbLineStringZM = 3002,         /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbPolygonZM = 3003,            /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiPointZM = 3004,         /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiLineStringZM = 3005,    /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiPolygonZM = 3006,       /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbGeometryCollectionZM = 3007, /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCircularStringZM = 3008,     /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCompoundCurveZM = 3009,      /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCurvePolygonZM = 3010,       /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiCurveZM = 3011,         /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbMultiSurfaceZM = 3012,       /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbCurveZM = 3013,              /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbSurfaceZM = 3014,            /**< ISO SQL/MM Part 3. GDAL &gt;= 2.1 */
+    wkbPolyhedralSurfaceZM = 3015,  /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTINZM = 3016,                /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
+    wkbTriangleZM = 3017,           /**< ISO SQL/MM Part 3. Reserved in GDAL &gt;= 2.1 but not yet implemented */
 
     wkbPoint25D = 0x80000001, /**< 2.5D extension as per 99-402 */
     wkbLineString25D = 0x80000002, /**< 2.5D extension as per 99-402 */
@@ -350,15 +406,11 @@ typedef enum
 
 } OGRwkbGeometryType;
 
-/* Outside of OGRwkbGeometryType since they are abstract types */
-#define wkbCurve            ((OGRwkbGeometryType)13)      /**< Curve (abstract type). SF-SQL 1.2 */
-#define wkbSurface          ((OGRwkbGeometryType)14)      /**< Surface (abstract type). SF-SQL 1.2 */
-
 /**
- * Output variants of WKB we support. 
+ * Output variants of WKB we support.
  *
  * 99-402 was a short-lived extension to SFSQL 1.1 that used a high-bit flag
- * to indicate the presence of Z coordiantes in a WKB geometry.
+ * to indicate the presence of Z coordinates in a WKB geometry.
  *
  * SQL/MM Part 3 and SFSQL 1.2 use offsets of 1000 (Z), 2000 (M) and 3000 (ZM)
  * to indicate the present of higher dimensional coordinates in a WKB geometry.
@@ -368,33 +420,49 @@ typedef enum
  * The codes are also found in §8.2.3 of <a href="http://portal.opengeospatial.org/files/?artifact_id=25355">
  * OGC 06-103r4 "OpenGIS® Implementation Standard for Geographic information - Simple feature access - Part 1: Common architecture", v1.2.1</a>
  */
-typedef enum 
+typedef enum
 {
     wkbVariantOldOgc, /**< Old-style 99-402 extended dimension (Z) WKB types */
     wkbVariantIso,  /**< SFSQL 1.2 and ISO SQL/MM Part 3 extended dimension (Z&M) WKB types */
     wkbVariantPostGIS1 /**< PostGIS 1.X has different codes for CurvePolygon, MultiCurve and MultiSurface */
 } OGRwkbVariant;
 
-
-/** @deprecated in GDAL 2.0. Use wkbHasZ() or wkbSetZ() instead */
 #ifndef GDAL_COMPILATION
+/** @deprecated in GDAL 2.0. Use wkbHasZ() or wkbSetZ() instead */
 #define wkb25DBit 0x80000000
 #endif
 
+#ifndef __cplusplus
 /** Return the 2D geometry type corresponding to the specified geometry type */
 #define wkbFlatten(x)  OGR_GT_Flatten((OGRwkbGeometryType)(x))
+#else
+/** Return the 2D geometry type corresponding to the specified geometry type */
+#define wkbFlatten(x)  OGR_GT_Flatten(static_cast<OGRwkbGeometryType>(x))
+#endif
 
 /** Return if the geometry type is a 3D geometry type
   * @since GDAL 2.0
   */
-#define wkbHasZ(x)     OGR_GT_HasZ(x)
+#define wkbHasZ(x)     (OGR_GT_HasZ(x) != 0)
 
 /** Return the 3D geometry type corresponding to the specified geometry type.
   * @since GDAL 2.0
   */
 #define wkbSetZ(x)     OGR_GT_SetZ(x)
 
+/** Return if the geometry type is a measured geometry type
+  * @since GDAL 2.1
+  */
+#define wkbHasM(x)     (OGR_GT_HasM(x) != 0)
+
+/** Return the measured geometry type corresponding to the specified geometry type.
+  * @since GDAL 2.1
+  */
+#define wkbSetM(x)     OGR_GT_SetM(x)
+
+#ifndef DOXYGEN_SKIP
 #define ogrZMarker 0x21125711
+#endif
 
 const char CPL_DLL * OGRGeometryTypeToName( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGRMergeGeometryTypes( OGRwkbGeometryType eMain,
@@ -404,8 +472,10 @@ OGRwkbGeometryType CPL_DLL OGRMergeGeometryTypesEx( OGRwkbGeometryType eMain,
                                                     int bAllowPromotingToCurves );
 OGRwkbGeometryType CPL_DLL OGR_GT_Flatten( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGR_GT_SetZ( OGRwkbGeometryType eType );
+OGRwkbGeometryType CPL_DLL OGR_GT_SetM( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGR_GT_SetModifier( OGRwkbGeometryType eType, int bSetZ, int bSetM );
 int                CPL_DLL OGR_GT_HasZ( OGRwkbGeometryType eType );
+int                CPL_DLL OGR_GT_HasM( OGRwkbGeometryType eType );
 int                CPL_DLL OGR_GT_IsSubClassOf( OGRwkbGeometryType eType,
                                                 OGRwkbGeometryType eSuperType );
 int                CPL_DLL OGR_GT_IsCurve( OGRwkbGeometryType );
@@ -415,24 +485,29 @@ OGRwkbGeometryType CPL_DLL OGR_GT_GetCollection( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGR_GT_GetCurve( OGRwkbGeometryType eType );
 OGRwkbGeometryType CPL_DLL OGR_GT_GetLinear( OGRwkbGeometryType eType );
 
-typedef enum 
+/** Enumeration to describe byte order */
+typedef enum
 {
-    wkbXDR = 0,         /* MSB/Sun/Motoroloa: Most Significant Byte First   */
-    wkbNDR = 1          /* LSB/Intel/Vax: Least Significant Byte First      */
+    wkbXDR = 0,         /**< MSB/Sun/Motoroloa: Most Significant Byte First   */
+    wkbNDR = 1          /**< LSB/Intel/Vax: Least Significant Byte First      */
 } OGRwkbByteOrder;
 
+#ifndef DOXYGEN_SKIP
+
 #ifndef NO_HACK_FOR_IBM_DB2_V72
 #  define HACK_FOR_IBM_DB2_V72
 #endif
 
 #ifdef HACK_FOR_IBM_DB2_V72
-#  define DB2_V72_FIX_BYTE_ORDER(x) ((((x) & 0x31) == (x)) ? (OGRwkbByteOrder) ((x) & 0x1) : (x))
-#  define DB2_V72_UNFIX_BYTE_ORDER(x) ((unsigned char) (OGRGeometry::bGenerate_DB2_V72_BYTE_ORDER ? ((x) | 0x30) : (x)))
+#  define DB2_V72_FIX_BYTE_ORDER(x) ((((x) & 0x31) == (x)) ? ((x) & 0x1) : (x))
+#  define DB2_V72_UNFIX_BYTE_ORDER(x) CPL_STATIC_CAST(unsigned char, OGRGeometry::bGenerate_DB2_V72_BYTE_ORDER ? ((x) | 0x30) : (x))
 #else
 #  define DB2_V72_FIX_BYTE_ORDER(x) (x)
 #  define DB2_V72_UNFIX_BYTE_ORDER(x) (x)
 #endif
 
+#endif /* #ifndef DOXYGEN_SKIP */
+
 /** Alter field name.
  * Used by OGR_L_AlterFieldDefn().
  */
@@ -465,7 +540,6 @@ typedef enum
  */
 #define ALTER_ALL_FLAG             (ALTER_NAME_FLAG | ALTER_TYPE_FLAG | ALTER_WIDTH_PRECISION_FLAG | ALTER_NULLABLE_FLAG | ALTER_DEFAULT_FLAG)
 
-
 /** Validate that fields respect not-null constraints.
  * Used by OGR_F_Validate().
  * @since GDAL 2.0
@@ -493,11 +567,19 @@ typedef enum
  */
 #define OGR_F_VAL_ALLOW_NULL_WHEN_DEFAULT       0x00000008
 
-/** Enable all validation tests.
+/** Allow geometry fields to have a different coordinate dimension that their
+ * geometry column type.
+ * This flag only makes sense if OGR_F_VAL_GEOM_TYPE is set too.
+ * Used by OGR_F_Validate().
+ * @since GDAL 2.1
+ */
+#define OGR_F_VAL_ALLOW_DIFFERENT_GEOM_DIM       0x00000010
+
+/** Enable all validation tests (except OGR_F_VAL_ALLOW_DIFFERENT_GEOM_DIM)
  * Used by OGR_F_Validate().
  * @since GDAL 2.0
  */
-#define OGR_F_VAL_ALL            0xFFFFFFFF
+#define OGR_F_VAL_ALL            (0x7FFFFFFF & ~OGR_F_VAL_ALLOW_DIFFERENT_GEOM_DIM)
 
 /************************************************************************/
 /*                  ogr_feature.h related definitions.                  */
@@ -509,7 +591,7 @@ typedef enum
  * field types can be known.
  */
 
-typedef enum 
+typedef enum
 {
   /** Simple 32bit integer */                   OFTInteger = 0,
   /** List of 32bit integers */                 OFTIntegerList = 1,
@@ -544,7 +626,7 @@ typedef enum
                                                         OFSTBoolean = 1,
     /** Signed 16-bit integer. Only valid for OFTInteger and OFTIntegerList. */
                                                         OFSTInt16 = 2,
-    /** Single precision (32 bit) floatint point. Only valid for OFTReal and OFTRealList. */
+    /** Single precision (32 bit) floating point. Only valid for OFTReal and OFTRealList. */
                                                         OFSTFloat32 = 3,
                                                         OFSTMaxSubType = 3
 } OGRFieldSubType;
@@ -553,16 +635,39 @@ typedef enum
  * Display justification for field values.
  */
 
-typedef enum 
+typedef enum
 {
     OJUndefined = 0,
     OJLeft = 1,
     OJRight = 2
 } OGRJustification;
 
+/** Special value for a unset FID */
 #define OGRNullFID            -1
+
+/* Special value for an unknown field type. This should only be used
+ * while reading a file. At the end of file any unknown types should
+ * be set to OFTString.
+*/
+/*! @cond Doxygen_Suppress */
+#define OGRUnknownType        static_cast<OGRFieldType>(-1)
+/*! @endcond */
+
+/** Special value set in OGRField.Set.nMarker1, nMarker2 and nMarker3 for
+ *  a unset field.
+ *  Direct use of this value is strongly discouraged.
+ *  Use OGR_RawField_SetUnset() or OGR_RawField_IsUnset() instead.
+ */
 #define OGRUnsetMarker        -21121
 
+/** Special value set in OGRField.Set.nMarker1, nMarker2 and nMarker3 for
+ *  a null field.
+ *  Direct use of this value is strongly discouraged.
+ *  Use OGR_RawField_SetNull() or OGR_RawField_IsNull() instead.
+ *  @since GDAL 2.2
+ */
+#define OGRNullMarker         -21122
+
 /************************************************************************/
 /*                               OGRField                               */
 /************************************************************************/
@@ -572,16 +677,17 @@ typedef enum
  */
 
 typedef union {
+/*! @cond Doxygen_Suppress */
     int         Integer;
     GIntBig     Integer64;
     double      Real;
     char       *String;
-    
+
     struct {
         int     nCount;
         int     *paList;
     } IntegerList;
-    
+
     struct {
         int     nCount;
         GIntBig *paList;
@@ -591,7 +697,7 @@ typedef union {
         int     nCount;
         double  *paList;
     } RealList;
-    
+
     struct {
         int     nCount;
         char    **paList;
@@ -601,10 +707,11 @@ typedef union {
         int     nCount;
         GByte   *paData;
     } Binary;
-    
+
     struct {
         int     nMarker1;
         int     nMarker2;
+        int     nMarker3;
     } Set;
 
     struct {
@@ -613,48 +720,62 @@ typedef union {
         GByte   Day;
         GByte   Hour;
         GByte   Minute;
-        GByte   TZFlag; /* 0=unknown, 1=localtime(ambiguous), 
+        GByte   TZFlag; /* 0=unknown, 1=localtime(ambiguous),
                            100=GMT, 104=GMT+1, 80=GMT-5, etc */
         GByte   Reserved; /* must be set to 0 */
         float   Second; /* with millisecond accuracy. at the end of the structure, so as to keep it 12 bytes on 32 bit */
     } Date;
+/*! @endcond */
 } OGRField;
 
-#define OGR_GET_MS(floatingpoint_sec)   (int)(((floatingpoint_sec) - (int)(floatingpoint_sec)) * 1000 + 0.5)
+#ifdef __cplusplus
+/** Return the number of milliseconds from a datetime with decimal seconds */
+inline int OGR_GET_MS(float fSec) {
+  if( CPLIsNan(fSec) ) return 0;
+  if( fSec >= 999 ) return 999;
+  if( fSec <= 0 ) return 0;
+  const float fValue = (fSec - static_cast<int>(fSec)) * 1000 + 0.5f;
+  return static_cast<int>(fValue);
+}
+#endif  // __cplusplus
 
-int CPL_DLL OGRParseDate( const char *pszInput, OGRField *psOutput, 
+int CPL_DLL OGRParseDate( const char *pszInput, OGRField *psOutput,
                           int nOptions );
 
 /* -------------------------------------------------------------------- */
 /*      Constants from ogrsf_frmts.h for capabilities.                  */
 /* -------------------------------------------------------------------- */
-#define OLCRandomRead          "RandomRead"
-#define OLCSequentialWrite     "SequentialWrite"
-#define OLCRandomWrite         "RandomWrite"
-#define OLCFastSpatialFilter   "FastSpatialFilter"
-#define OLCFastFeatureCount    "FastFeatureCount"
-#define OLCFastGetExtent       "FastGetExtent"
-#define OLCCreateField         "CreateField"
-#define OLCDeleteField         "DeleteField"
-#define OLCReorderFields       "ReorderFields"
-#define OLCAlterFieldDefn      "AlterFieldDefn"
-#define OLCTransactions        "Transactions"
-#define OLCDeleteFeature       "DeleteFeature"
-#define OLCFastSetNextByIndex  "FastSetNextByIndex"
-#define OLCStringsAsUTF8       "StringsAsUTF8"
-#define OLCIgnoreFields        "IgnoreFields"
-#define OLCCreateGeomField     "CreateGeomField"
-#define OLCCurveGeometries     "CurveGeometries"
+#define OLCRandomRead          "RandomRead"         /**< Layer capability for random read */
+#define OLCSequentialWrite     "SequentialWrite"    /**< Layer capability for sequential write */
+#define OLCRandomWrite         "RandomWrite"        /**< Layer capability for random write */
+#define OLCFastSpatialFilter   "FastSpatialFilter"  /**< Layer capability for fast spatial filter */
+#define OLCFastFeatureCount    "FastFeatureCount"   /**< Layer capability for fast feature count retrieval */
+#define OLCFastGetExtent       "FastGetExtent"      /**< Layer capability for fast extent retrieval */
+#define OLCCreateField         "CreateField"        /**< Layer capability for field creation */
+#define OLCDeleteField         "DeleteField"        /**< Layer capability for field deletion */
+#define OLCReorderFields       "ReorderFields"      /**< Layer capability for field reordering */
+#define OLCAlterFieldDefn      "AlterFieldDefn"     /**< Layer capability for field alteration */
+#define OLCTransactions        "Transactions"       /**< Layer capability for transactions */
+#define OLCDeleteFeature       "DeleteFeature"      /**< Layer capability for feature deletion */
+#define OLCFastSetNextByIndex  "FastSetNextByIndex" /**< Layer capability for setting next feature index */
+#define OLCStringsAsUTF8       "StringsAsUTF8"      /**< Layer capability for strings returned with UTF-8 encoding */
+#define OLCIgnoreFields        "IgnoreFields"       /**< Layer capability for field ignoring */
+#define OLCCreateGeomField     "CreateGeomField"    /**< Layer capability for geometry field creation */
+#define OLCCurveGeometries     "CurveGeometries"    /**< Layer capability for curve geometries support */
+#define OLCMeasuredGeometries  "MeasuredGeometries" /**< Layer capability for measured geometries support */
 
-#define ODsCCreateLayer        "CreateLayer"
-#define ODsCDeleteLayer        "DeleteLayer"
-#define ODsCCreateGeomFieldAfterCreateLayer   "CreateGeomFieldAfterCreateLayer"
-#define ODsCCurveGeometries    "CurveGeometries"
-#define ODsCTransactions       "Transactions"
-#define ODsCEmulatedTransactions "EmulatedTransactions"
+#define ODsCCreateLayer        "CreateLayer"        /**< Dataset capability for layer creation */
+#define ODsCDeleteLayer        "DeleteLayer"        /**< Dataset capability for layer deletion */
+#define ODsCCreateGeomFieldAfterCreateLayer   "CreateGeomFieldAfterCreateLayer" /**< Dataset capability for geometry field creation support */
+#define ODsCCurveGeometries    "CurveGeometries"    /**< Dataset capability for curve geometries support */
+#define ODsCTransactions       "Transactions"       /**< Dataset capability for dataset transcations */
+#define ODsCEmulatedTransactions "EmulatedTransactions" /**< Dataset capability for emulated dataset transactions */
+#define ODsCMeasuredGeometries "MeasuredGeometries"     /**< Dataset capability for measured geometries support */
+#define ODsCRandomLayerRead     "RandomLayerRead"   /**< Dataset capability for GetNextFeature() returning features from random layers */
+#define ODsCRandomLayerWrite    "RandomLayerWrite " /**< Dataset capability for supporting CreateFeature on layer in random order */
 
-#define ODrCCreateDataSource   "CreateDataSource"
-#define ODrCDeleteDataSource   "DeleteDataSource"
+#define ODrCCreateDataSource   "CreateDataSource"   /**< Driver capability for datasource creation */
+#define ODrCDeleteDataSource   "DeleteDataSource"   /**< Driver capability for datasource deletion */
 
 /* -------------------------------------------------------------------- */
 /*      Layer metadata items.                                           */
@@ -675,12 +796,12 @@ int CPL_DLL OGRParseDate( const char *pszInput, OGRField *psOutput,
 
 typedef enum ogr_style_tool_class_id
 {
-    OGRSTCNone   = 0,
-    OGRSTCPen    = 1,
-    OGRSTCBrush  = 2,
-    OGRSTCSymbol = 3,
-    OGRSTCLabel  = 4,
-    OGRSTCVector = 5
+    OGRSTCNone   = 0, /**< None */
+    OGRSTCPen    = 1, /**< Pen */
+    OGRSTCBrush  = 2, /**< Brush */
+    OGRSTCSymbol = 3, /**< Symbol */
+    OGRSTCLabel  = 4, /**< Label */
+    OGRSTCVector = 5  /**< Vector */
 } OGRSTClassId;
 
 /**
@@ -688,104 +809,110 @@ typedef enum ogr_style_tool_class_id
  */
 typedef enum ogr_style_tool_units_id
 {
-    OGRSTUGround = 0,
-    OGRSTUPixel  = 1,
-    OGRSTUPoints = 2,
-    OGRSTUMM     = 3,
-    OGRSTUCM     = 4,
-    OGRSTUInches = 5
+    OGRSTUGround = 0, /**< Ground unit */
+    OGRSTUPixel  = 1, /**< Pixel */
+    OGRSTUPoints = 2, /**< Points */
+    OGRSTUMM     = 3, /**< Millimeter */
+    OGRSTUCM     = 4, /**< Centimeter */
+    OGRSTUInches = 5  /**< Inch */
 } OGRSTUnitId;
 
 /**
  * List of parameters for use with OGRStylePen.
  */
 typedef enum ogr_style_tool_param_pen_id
-{  
-    OGRSTPenColor       = 0,                   
-    OGRSTPenWidth       = 1,                   
-    OGRSTPenPattern     = 2,
-    OGRSTPenId          = 3,
-    OGRSTPenPerOffset   = 4,
-    OGRSTPenCap         = 5,
-    OGRSTPenJoin        = 6,
-    OGRSTPenPriority    = 7,
+{
+    OGRSTPenColor       = 0, /**< Color */
+    OGRSTPenWidth       = 1, /**< Width */
+    OGRSTPenPattern     = 2, /**< Pattern */
+    OGRSTPenId          = 3, /**< Id */
+    OGRSTPenPerOffset   = 4, /**< Perpendicular offset */
+    OGRSTPenCap         = 5, /**< Cap */
+    OGRSTPenJoin        = 6, /**< Join */
+    OGRSTPenPriority    = 7, /**< Priority */
+#ifndef DOXYGEN_SKIP
     OGRSTPenLast        = 8
-              
+#endif
 } OGRSTPenParam;
 
 /**
  * List of parameters for use with OGRStyleBrush.
  */
 typedef enum ogr_style_tool_param_brush_id
-{  
-    OGRSTBrushFColor    = 0,                   
-    OGRSTBrushBColor    = 1,                   
-    OGRSTBrushId        = 2,
-    OGRSTBrushAngle     = 3,                   
-    OGRSTBrushSize      = 4,
-    OGRSTBrushDx        = 5,
-    OGRSTBrushDy        = 6,
-    OGRSTBrushPriority  = 7,
+{
+    OGRSTBrushFColor    = 0, /**< Foreground color */
+    OGRSTBrushBColor    = 1, /**< Background color */
+    OGRSTBrushId        = 2, /**< Id */
+    OGRSTBrushAngle     = 3, /**< Angle */
+    OGRSTBrushSize      = 4, /**< Size */
+    OGRSTBrushDx        = 5, /**< Dx */
+    OGRSTBrushDy        = 6, /**< Dy */
+    OGRSTBrushPriority  = 7, /**< Priority */
+#ifndef DOXYGEN_SKIP
     OGRSTBrushLast      = 8
-              
-} OGRSTBrushParam;
+#endif
 
+} OGRSTBrushParam;
 
 /**
  * List of parameters for use with OGRStyleSymbol.
  */
 typedef enum ogr_style_tool_param_symbol_id
-{  
-    OGRSTSymbolId       = 0,
-    OGRSTSymbolAngle    = 1,
-    OGRSTSymbolColor    = 2,
-    OGRSTSymbolSize     = 3,
-    OGRSTSymbolDx       = 4,
-    OGRSTSymbolDy       = 5,
-    OGRSTSymbolStep     = 6,
-    OGRSTSymbolPerp     = 7,
-    OGRSTSymbolOffset   = 8,
-    OGRSTSymbolPriority = 9,
-    OGRSTSymbolFontName = 10,
-    OGRSTSymbolOColor   = 11,
+{
+    OGRSTSymbolId       = 0, /**< Id */
+    OGRSTSymbolAngle    = 1, /**< Angle */
+    OGRSTSymbolColor    = 2, /**< Color */
+    OGRSTSymbolSize     = 3, /**< Size */
+    OGRSTSymbolDx       = 4, /**< Dx */
+    OGRSTSymbolDy       = 5, /**< Dy */
+    OGRSTSymbolStep     = 6, /**< Step */
+    OGRSTSymbolPerp     = 7, /**< Perpendicular */
+    OGRSTSymbolOffset   = 8, /**< Offset */
+    OGRSTSymbolPriority = 9, /**< Priority */
+    OGRSTSymbolFontName = 10, /**< OBSOLETE; do not use */
+    OGRSTSymbolOColor   = 11, /**< Outline color */
+#ifndef DOXYGEN_SKIP
     OGRSTSymbolLast     = 12
-              
+#endif
 } OGRSTSymbolParam;
 
 /**
  * List of parameters for use with OGRStyleLabel.
  */
 typedef enum ogr_style_tool_param_label_id
-{  
-    OGRSTLabelFontName  = 0,
-    OGRSTLabelSize      = 1,
-    OGRSTLabelTextString = 2,
-    OGRSTLabelAngle     = 3,
-    OGRSTLabelFColor    = 4,
-    OGRSTLabelBColor    = 5,
-    OGRSTLabelPlacement = 6,
-    OGRSTLabelAnchor    = 7,
-    OGRSTLabelDx        = 8,
-    OGRSTLabelDy        = 9,
-    OGRSTLabelPerp      = 10,
-    OGRSTLabelBold      = 11,
-    OGRSTLabelItalic    = 12,
-    OGRSTLabelUnderline = 13,
-    OGRSTLabelPriority  = 14,
-    OGRSTLabelStrikeout = 15,
-    OGRSTLabelStretch   = 16,
-    OGRSTLabelAdjHor    = 17,
-    OGRSTLabelAdjVert   = 18,
-    OGRSTLabelHColor    = 19,
-    OGRSTLabelOColor    = 20,
+{
+    OGRSTLabelFontName  = 0, /**< Font name */
+    OGRSTLabelSize      = 1, /**< Size */
+    OGRSTLabelTextString = 2, /**< Text string */
+    OGRSTLabelAngle     = 3, /**< Angle */
+    OGRSTLabelFColor    = 4, /**< Foreground color */
+    OGRSTLabelBColor    = 5, /**< Background color */
+    OGRSTLabelPlacement = 6, /**< Placement */
+    OGRSTLabelAnchor    = 7, /**< Anchor */
+    OGRSTLabelDx        = 8, /**< Dx */
+    OGRSTLabelDy        = 9, /**< Dy */
+    OGRSTLabelPerp      = 10, /**< Perpendicular */
+    OGRSTLabelBold      = 11, /**< Bold */
+    OGRSTLabelItalic    = 12, /**< Italic */
+    OGRSTLabelUnderline = 13, /**< Underline */
+    OGRSTLabelPriority  = 14, /**< Priority */
+    OGRSTLabelStrikeout = 15, /**< Strike out */
+    OGRSTLabelStretch   = 16, /**< Stretch */
+    OGRSTLabelAdjHor    = 17, /**< OBSOLETE; do not use */
+    OGRSTLabelAdjVert   = 18, /**< OBSOLETE; do not use */
+    OGRSTLabelHColor    = 19, /**< Highlight color */
+    OGRSTLabelOColor    = 20, /**< Outline color */
+#ifndef DOXYGEN_SKIP
     OGRSTLabelLast      = 21
-              
+#endif
 } OGRSTLabelParam;
 
 /* ------------------------------------------------------------------- */
 /*                        Version checking                             */
 /* -------------------------------------------------------------------- */
 
+#ifndef DOXYGEN_SKIP
+
 /* Note to developers : please keep this section in sync with gdal.h */
 
 #ifndef GDAL_VERSION_INFO_DEFINED
@@ -798,12 +925,12 @@ const char CPL_DLL * CPL_STDCALL GDALVersionInfo( const char * );
 /** Return TRUE if GDAL library version at runtime matches nVersionMajor.nVersionMinor.
 
     The purpose of this method is to ensure that calling code will run with the GDAL
-    version it is compiled for. It is primarly intented for external plugins.
+    version it is compiled for. It is primarily indented for external plugins.
 
     @param nVersionMajor Major version to be tested against
     @param nVersionMinor Minor version to be tested against
     @param pszCallingComponentName If not NULL, in case of version mismatch, the method
-                                   will issue a failure mentionning the name of
+                                   will issue a failure mentioning the name of
                                    the calling component.
   */
 int CPL_DLL CPL_STDCALL GDALCheckVersion( int nVersionMajor, int nVersionMinor,
@@ -815,6 +942,8 @@ int CPL_DLL CPL_STDCALL GDALCheckVersion( int nVersionMajor, int nVersionMinor,
 
 #endif
 
+#endif /* #ifndef DOXYGEN_SKIP */
+
 CPL_C_END
 
 #endif /* ndef OGR_CORE_H_INCLUDED */
